// Welcome to
// __________         __    __  .__                               __
// \______   \_____ _/  |__/  |_|  |   ____   ______ ____ _____  |  | __ ____
//  |    |  _/\__  \\   __\   __\  | _/ __ \ /  ___//    \\__  \ |  |/ // __ \
//  |    |   \ / __ \|  |  |  | |  |_\  ___/ \___ \|   |  \/ __ \|    <\  ___/
//  |________/(______/__|  |__| |____/\_____>______>___|__(______/__|__\\_____>
//
// This file can be a nice home for your Battlesnake logic and helper functions.
//
// To get you started we've included code to prevent your Battlesnake from moving backwards.
// For more info see docs.battlesnake.com

use log::info;
use rand::seq::SliceRandom;
use serde_json::{json, Value};
use std::collections::HashMap;

use crate::{Battlesnake, Board, Coord, Direction, Game, Move};

// info is called when you create your Battlesnake on play.battlesnake.com
// and controls your Battlesnake's appearance
// TIP: If you open your Battlesnake URL in a browser you should see this data
pub fn info() -> Value {
    info!("INFO");

    json!({
        "apiversion": "1",
        "author": "der-informatiker",
        "color": "#00FFEE",
        "head": "smart-caterpillar",
        "tail": "mouse",
    })
}

// start is called when your Battlesnake begins a game
pub fn start(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME START");
}

// end is called when your Battlesnake finishes a game
pub fn end(_game: &Game, _turn: i32, _board: &Board, _you: &Battlesnake) {
    info!("GAME OVER");
}

// move is called on every turn and returns your next move
// Valid moves are "up", "down", "left", or "right"
// See https://docs.battlesnake.com/api/example-move for available data
pub fn get_move(_game: &Game, turn: i32, board: &Board, you: &Battlesnake) -> Move {
    use Direction::{Down, Left, Right, Up};

    let mut is_move_safe: HashMap<_, _> =
        vec![(Up, true), (Down, true), (Left, true), (Right, true)]
            .into_iter()
            .collect();

    // We've included code to prevent your Battlesnake from moving backwards
    let my_head = &you.body[0]; // Coordinates of your head
    let my_neck = &you.body[1]; // Coordinates of your "neck"

    if my_neck.x < my_head.x {
        // Neck is left of head, don't move left
        is_move_safe.insert(Left, false);
    } else if my_neck.x > my_head.x {
        // Neck is right of head, don't move right
        is_move_safe.insert(Right, false);
    } else if my_neck.y < my_head.y {
        // Neck is below head, don't move down
        is_move_safe.insert(Down, false);
    } else if my_neck.y > my_head.y {
        // Neck is above head, don't move up
        is_move_safe.insert(Up, false);
    }

    // TODO: Step 1 - Prevent your Battlesnake from moving out of bounds
    if my_head.x == 0 {
        is_move_safe.insert(Left, false);
    }
    if my_head.x == board.width - 1 {
        is_move_safe.insert(Right, false);
    }
    if my_head.y == 0 {
        is_move_safe.insert(Down, false);
    }
    if my_head.y == board.height - 1 {
        is_move_safe.insert(Up, false);
    }

    // TODO: Step 2 - Prevent your Battlesnake from colliding with itself
    if you.body.contains(&Coord {
        x: my_head.x - 1,
        y: my_head.y,
    }) {
        is_move_safe.insert(Left, false);
    }
    if you.body.contains(&Coord {
        x: my_head.x + 1,
        y: my_head.y,
    }) {
        is_move_safe.insert(Right, false);
    }
    if you.body.contains(&Coord {
        x: my_head.x,
        y: my_head.y - 1,
    }) {
        is_move_safe.insert(Down, false);
    }
    if you.body.contains(&Coord {
        x: my_head.x,
        y: my_head.y + 1,
    }) {
        is_move_safe.insert(Up, false);
    }

    // TODO: Step 3 - Prevent your Battlesnake from colliding with other Battlesnakes
    // let opponents = &board.snakes;

    // Are there any safe moves left?
    let safe_moves = is_move_safe
        .into_iter()
        .filter(|&(_, v)| v)
        .map(|(k, _)| k)
        .collect::<Vec<_>>();

    // Choose a random move from the safe ones
    let chosen = safe_moves.choose(&mut rand::thread_rng()).unwrap_or(&Up);

    // TODO: Step 4 - Move towards food instead of random, to regain health and survive longer
    // let food = &board.food;

    info!("MOVE {}: {:?}", turn, chosen);
    Move {
        r#move: *chosen,
        shout: None,
    }
}
